
from camagick.processor import ProcessorBase

from asteval import Interpreter

def populate_argparser(parser):
    pass

class Processor(ProcessorBase):
    '''
    Changes the array shape of the input data.
    '''
    
    def __init__(self, *dims, shape=None, **kw):
        '''
        Args:
            dims: List of shape specification, one per dimension. If the dimension
              specification are strings, they are parsed using `asteval.Interpreter`
              from the `lmfit` people, i.e. may contain a limited set of mathematical
              functions, constants, and data context tags as symbols.

            shape: Alternatively, if there are no dims, a shape can be supplied
              as a whole. As string, this is also parsed using `asteval.Interperter()`.
        
        '''
        super().__init__(**kw)
        if len(dims)==0:
            self.shape = shape
        else:
            self.shape = dims

    async def __call__(self, data=None, context=None):
        #shape = (context.get(f'{k}_shape', self.shape) for k in data) \
        #    if context is not None else self.shape

        if self.shape is None:
            raise RuntimeError(f'msg="{self.__class__.__name__} requires a shape '
                               f'for {k} got none"')
        
        
        inter = Interpreter()
        inter.symtable.update(context or {})
        
        if isinstance(self.shape, str):
            shape = inter(self.shape)
            if shape is None:
                raise RuntimeError(f'msg="Cannot parse shape" spec="{self.shape}"')
        else:
            shape = tuple([
                (inter(s, raise_errors=True) \
                 if isinstance(s, str) else s)  for s in self.shape
            ])
            for s1,s2 in zip(shape, self.shape):
                if s1 is None:
                    raise RuntimeError(f'msg="Don\'t understand shape" '
                                       f'spec="{self.shape}" '
                                       f'result="{shape}"')
        
        return { k:v.reshape(shape) for k,v in data.items() }



#!/usr/bin/python3

from numpy import zeros, sqrt
import math, time

import asteval

class ArgEval:
    '''
    Helper to transform (a list of) string-based arguments into
    pythonic data types (floats, ints, lists, dict, ...) while
    applying a mild form of expression evaluation based on
    `lmfit`'s `asteval`.

    The "source code" is loaded from a single line. It's expected
    to either be a single expression, or a short-form of a Python
    dictionary of the form "foo:..,bar:..." (representing the
    long form "{'foo': ..., 'bar': ...}").
    '''
    def __init__(self, line: str, *keys, **kfmt):
        '''
        Args:
          line: The source code
          *keys: list of strings representing possible keys in
            dictionaries. For every "key" a corresponding symbol
            by the same name, and with the string as a content,
            will be preset in the `asteval` interpreter.
          **kfmt: Alternative way of specifying the keys. Each
            key of `kfmt` will be treated as one of the `*keys`,
            i.e. a corresponding symbol will be created in the
            interpreter; 
        '''
        pass
        

class FpsCounter:
    def __init__(self, pts=100, label="FPS"):
        self.told = time.time()
        self.told_diff = 0
        self.fps_history = zeros(pts)
        self.ji = 0
        self.label = label
        
    def mark(self):
        tnow = time.time()
        tdiff = 1/(tnow-self.told)
        self.fps_history[self.ji] = tdiff
        self.ji = (self.ji+1)%self.fps_history.shape[0]
        self.told = tnow

    @property
    def fps(self):
        return self.fps_history.mean()

    @property
    def jitter(self):
        return sqrt( ((self.fps_history-self.fps_history.mean())**2).mean() )

    def __str__(self):
        return "%s: %2.1f ± %d Hz" % (self.label, self.fps, math.ceil(self.jitter))


    def __repr__(self):
        return str(self)



from camagick.processor import ProcessorBase

import logging

from asteval import Interpreter

class Processor(ProcessorBase):
    '''
    Executes arithmetic computations on the input data.
    '''    
    def __init__(self, expr):
        '''
        Args:
            expr: Python-esque expression to calculate. "{}" is being
              substituted for the input data, any other symbol for the
              corresponding context field. The expression is evaluated
              using lmfit's `asteval` module.
        '''
        self._expr = expr.format('__input_data__')
        self._intrp = Interpreter()
        
    
    async def startup(self):
        pass

    async def shutdown(self):
        pass
    
    async def __call__(self, data=None, context=None):
        intrp = self._intrp
        intrp.symtable.update(**context)
        ret = {}
        for k,v in data.items():
            intrp.symtable['__input_data__'] = v
            ret[k] = intrp(self._expr)
        return ret



from camagick.storage.base import ArrayAccess as ArrayAccessBase
from camagick.storage.base import RefuseIndex, RefuseOverwrite, RefuseSparse

import logging

logger = logging.getLogger(__name__)    

from contextlib import contextmanager 

class ArrayAccess(ArrayAccessBase):
    '''
    Implementation of a Zarr array access.
    '''

    def __enter__(self):
        import contextlib, zarr

        #self._outer_object = self._exit_stack.enter_context(
        #self._outer_object = zarr.open(store=self.outer_path, mode=('r' if self.readonly else 'a'))
        self._outer_object = zarr.open(store=self.outer_path)

        #self._exit_stack = contextlib.ExitStack()        
        #self._exit_stack.enter_context(self._outer_object.store.fs.transaction)
            
        return self


    def outer_object(self, tag):
        # Ignore the tag -- this backend currently only supports single backend
        return self._outer_object
    

    def __exit__(self, *args):
        #self._exit_stack.close()
        pass


    def ensure_group(self, name, attrs=None):
        # Makes sure `name` exists in the current zarr store.
        # Possibly updates attributes.
        node = self.outer_object(None)
        if name not in node:
            obj = node.create_group(current)
        else:
            obj = node[name]

        if attrs is not None:
            obj.attrs.update(attrs)

        return obj


    def ensure_dataset(self, path, dtype=None, frame_shape=None, num_frames=0,
                       attrs=None):
        '''
        Received with the first data point, sets names, types etc.

        Args:
            path: name or path of the dataset        
            dtype: data type (numpy-dtype)
            frame_shape: shape of the data point; note that
              the actual shape of the resulting data set will contain
              more dimension
            attrs: attributes to set for the dataset object
        '''

        base = self.outer_object(None)

        if self.readonly:
            return base[path] if len(path) > 0 else base
        
        dshape = (0, *frame_shape) if frame_shape is not None else (0,)
        
        try:
            dset = base[path]

            if (dset.dtype != dtype):
                if  'O' not in self._frame_mode:
                    raise RuntimeError(f'msg="Dataset exists with different dtype"'
                                       f'have="{dset.dtype}" want="{dtype}"')
                else:
                    del base[path]
                    raise KeyError()

            if dset.shape[1:] != dshape[1:]:
                if  'O' not in self._frame_mode:                
                    raise RuntimeError(f'msg="Dataset exists with different shape"'
                                       f'have="{dset.shape}" want="{dshape}"')
                else:
                    del base[path]
                    raise KeyError()

        except KeyError:
            logger.info(f'msg="Creating dataset" file="{str(base.store)}" name="{path}" '
                        f'shape="{dshape}" dtype={dtype}')
            dset = base.create_array(path,
                                     shape=dshape,
                                     #maxshape=(None, *dshape[1:]),
                                     dtype=dtype)

        if attrs is not None:
            dset.attrs.update(attrs)

        return dset



'''
Array Storage API abstraction

This is meant to implement the operations that CAmagick supports in
an backend-agnostic way. Typically, CAmagick writes to HDF5, Zarr
and Tiled. So this API intents to support the common denominator of
all of these:

- Eternal / internal path (i.e. an entry point, "file" or "url"
  that designates a high-level resource, and an internal path
  that designates a low-level resource within the high-level
  one)

- Groups or Datasets (HDF5-like hiearchical structure, where
  there are datasets or groups, with the groups containing further
  datasets or subgroups, accessible by ["...path..."])

- Datasets can be sliced, and meta information (in particular data type
  and shape) can be accessed by `.dtype`, respectively `.shape`

- Datasets and/or groups can have attributes by means of `.attrs[...]`

- Files can be opened for reading, or for read-writing, optionally
  overwriting or refusing to create new files when opening for
  read-writing and the file already exists / doesn't exist.
'''

import time, logging
from numpy import array as np_array

logger = logging.getLogger(__name__)

__all__ = [
    "RefuseIndex",
    "RefuseOverwrite",
    "RefuseSparse",
    "ArrayAccess",
    "MemoryArrayAccess"
]

# list of procedures that return True if a specific storage
# path/url is of a specific format.

_format_probe = {
    'hdf5': lambda x: np_array([
        x.lower().endswith(s) for s in \
        ('.h5', '.hdf', '.hdf5', '.nx', '.cdf4')
    ]).any(),
    
    'zarr': lambda x: np_array([
        x.lower().endswith(s) for s in \
        ('.z', '.zarr')
    ]).any(),
}

class UnknownArrayType(RuntimeError): pass


def probe_url(url, backtrack=True):
    '''
    Tries to guess the "kind" of a data URL.

    Returns a named tuple (kind, outer, residual), where "kind" is a
    string that represents the type and is a submodule in
    `camagick.storage`, and "outer" is the outer-part of the URL
    that can be passed
    to any ArrayAccess object as a first parameter. "residual" is
    the remainder of the URL that didn't fit in "outer". This is
    in particular the case if the URL doesn't end with the
    storage format (e.g. "file:///path/to/file.hdf5/group/subgroup/").

    Probing formats is tricky business. There are 3 mechanisms in use,
    in this order:

      - Check the "scheme" part of the for something like "http+hdf5",
        where the part after "+" is the kind.

      - Check the extention of the final path element for known
        extensions, as in: "/path/to/file.hdf5", or possibly even
        "file:///path/to/file.hdf5#subgroup"

      - Backtracking the path elements one by one and check their
        extension, e.g. to find things like
        "/path/to/folder.zarr/subgroup/". This is only performed if
        `backtrack` is set to `True`.
    '''

    from urllib.parse import urlparse, urlunparse, ParseResult
    from collections import namedtuple

    purl = urlparse(url)

    UrlSpec = namedtuple('UrlSpec', ['kind', 'outer', 'residual'])

    # Step 1: check scheme (overrides any path element extensions)
    if purl.scheme not in (None, ''):
        tmp = purl.scheme.split('+')
        if len(tmp) == 2:
            return UrlSpec(tmp[1], url, '')

    # Steps 2+3: iteratively check file extensions of path elements
    path_elems = purl.path.split('/')
    nr_elems = len(path_elems)
    for j,elem in enumerate(path_elems[::-1]):
        i = nr_elems - j
        for kind, stest in _format_probe.items():
            if stest(elem):
                outer = urlunparse(ParseResult(
                    scheme=purl.scheme,
                    netloc=purl.netloc,
                    path='/'.join(path_elems[:i]),
                    params=purl.params,
                    query=purl.query,
                    fragment=None
                ))
                inner = '/'.join(path_elems[i:]) + \
                    (f'#{purl.fragment}' if purl.fragment not in (None, '') else '')
                # fragment=purl.fragment
                return UrlSpec(kind, outer, inner)

        if not backtrack:
            break

    raise UnknownArrayType(f'Unknown array type for: {url}')


def probe_kind(url):
    '''
    Wrapper for `probe_url_spec()` for backwards compatibility.
    Returns just the kind.
    '''
    return probe_url(url)[0]


def find_loader(url=None, kind=None):
    from importlib import import_module
    if kind is None:
        kind = probe_kind(url)
    mod = import_module(f'camagick.storage.{kind}')
    return getattr(mod, 'ArrayAccess')


from contextlib import contextmanager

@contextmanager
def load_from_url(url, kind=None, mode=None, **tags):
    '''
    Wrapper for find_loader(...) and the corresponding loading.
    '''
    Loader = find_loader(url, kind=kind)
    with Loader(url, mode, **tags) as obj:
        try:
            yield obj
        finally:
            pass


class RefuseFrameStorage(RuntimeError): pass

class RefuseShape(RefuseFrameStorage): pass

class RefuseIndex(RefuseFrameStorage): pass

class RefuseOverwrite(RefuseIndex): pass

class RefuseSparse(RefuseIndex): pass

class DatasetUnavailable(RuntimeError):
    def __init__(self, tag, reason=None):
        self.tag = tag
        super().__init__(f'Dataset unavailable: "{tag}" ({reason})')

class ArrayAccess:
    '''
    This is just to document the abstract API.

    The API elements are heavily tailored towards CAmagick's needs, which
    are essentially appending new data ("frames") to existing datasets.
    To implement a new backend, at least the following members need
    to be defined:

    - Store opening/closing: within context management infrastructure
      (`__enter__`, `__exit__`, and provide an `.outer_object` property)
    
    - Dataset management: `_ensure_dataset()`

    - (Optionally) dataset resizing: `._resize_dataset()` (and possibly
      `_reset_dataset()`), if the dataset doesn't have a `.resize(shape)`
      function.
  
    '''

    def __init__(self, outer=None, mode=None, **tags):
        '''
        Initializes the storage mechanism for "outer".

        We refer to resources in an hiearchical array storage by
        their URL. For every URL we differentiate between an
        "outer" part (prefix), pointing to the externally visible
        resource (e.g. HDF5 file), and an "inner" part, pointing
        to the resource within that file. Since we're focused on
        using a URL-like pattern, we're recycling the URL separator
        '#' for delimining outer from inner, so a typical URL might
        look like:
           "scheme://possibly.a.host.part/path/to/container#path/within/container"

        Here, everything up to '#' is the outer part, and everything
        after '#' is the inner part.

        One specific `ArrayAccess` implementation only covers one
        specific "outer" point -- this is sometimes important when
        the storage backend possibly locks, e.g. in the case of (writing)
        HDF5 files. This is the reason why we separate.

        We're liberal as to whether the repo URL is file#sub, or file/sub
        (maybe a little _too_ liberal...). As a consequence, here inner
        might be something like: "#...", or "/...", or just plain "...".
        But different backends react somewhat sensitive to that.
        We therefore give every backend the possibility to implement their
        own inner-path mangling by overriding the `._mangle_inner()` function.
        Mostly it's about detecting a '#', so this is the default implementation,
        but sometimes also checking whether there's an empty string involved
        (which HDF5 cannot handle) is important.
        
        Args:
            outer: Top-level store path (file name or URL).

            mode: What to do when we're required to create or overwrite a
              data point. We only concert ourselves with data points since
              this is what CAmagick's basic data currency is. For everything
              else (files, groups), the policy follows whatever must be done
              to ensure the data point mode policy can be followed.
        
              This must be either `None`, or a string consistinc or a
              combination of `a` (append), `o` or `O` (overwrite)
              or `x` (reset), possibly in combination with `+` (top-up),
              meaning:
        
                 - `a`: require that the data point be always appended
                   to the current dataset -- never overwrite or delete
                   existing data.
        
                 - `o`: overwrite already existing data points (frames).
                   If the shape of the new data point frames doesn't match
                   the old ones, then fail.

                 - `O`: same as `o`, but additioally delete and re-create
                   whole dataset entries, if the shape or data type of the
                   existing dataset does not match the new request.

                 - `x`: reset the entire dataset (i.e. delete all existing
                   data points) if a data point already exists at the
                   specified index.

                 - `+`: top-up marker will fill up empty slots, but without
                    actually writing any data to them. They will remain
                    filled with default settings of the corresponding backend
                    and the data type. If top-up is not specified, otherwise
                    the operation will fail if index is higher than the one of
                    the next slot in turn.

              If `mode` is set to `None`, each dataset, and consequently the
              whole store, is opened for reading only. This also reflects
              the behavior of the async environment enter/exit functionality.

            **tags: Keys are tag names (by which to refer the data in
              `.push_frame()`, and values are path elements, for each
              frame in itself.
        '''
        self._stor_spec = outer
        self._stor_dspath = {
            k:self._mangle_inner(v) for k,v in tags.items()
        }

        self._frame_mode = mode

        
    def _mangle_inner(self, inner):
        '''
        Mangle the inner path for an array storage resource (group or dataset).
        '''

        if len(inner) == 0:
            return '/'

        if inner[0] == '#':
            if len(inner) > 1:
                inner = inner[1:]
            else:
                inner = ''

        if len(inner) == 0:
            inner = '/'

        return inner


    def outer_obj(self, tag):
        raise RuntimeError('not implemented')


    @property
    def outer_path(self):
        '''
        Returns the "outer" part of the store specification.
        '''
        return self._stor_spec


    #@property
    #def inner_path(self):
    #    '''
    #    Returns the "inner" part of the store specification.
    #    '''
    #    return self._stor_spec[1]


    @property
    def readonly(self):
        return self._frame_mode is None


    # async wrappers for context management, which might block
    async def __aenter__(self):
        import concurrent, asyncio
        loop = asyncio.get_running_loop()
        with concurrent.futures.ThreadPoolExecutor() as pool:
            return await loop.run_in_executor(pool, self.__enter__)
        

    async def __aexit__(self, *args):
        import concurrent, asyncio
        loop = asyncio.get_running_loop()
        with concurrent.futures.ThreadPoolExecutor() as pool:
            return await loop.run_in_executor(pool, self.__exit__)

            
    def _split_dset_path(self, dset_path, splitter='/'):
        # Helper that will return the (node path, dataset path)
        # out of a whole path.
        # If there's only one component, assume it's a dataset.
        parts = dset_path.split(splitter)
        
        if len(parts) == 0:
            raise RuntimeError(f'invalid dataset path {dset_path}')

        if len(parts) == 1:
            return ('', parts[0])

        return ('/'.join(parts[:-1]), parts[-1])


    def ensure_group(self, name, attrs=None):
        '''
        Makse sure the group exists and has the proper attributes

        Args:
            name: The absolute group name within `outer_object`
        
            attrs: additional group attributes to set

        Returns a node object for the group.
        '''
        raise RuntimeError('not implemented')


    def ensure_dataset(self, path, dtype=None, frame_shape=None, num_frames=0,
                       attrs=None):
        '''
        Ensure the existence of the target dataset.
        
        This is typically called with the first received frame-push (i.e.
        data point), and is intended to set sets names, types,
        attributes etc.

        Args:
            path: full path of the dataset inside `base`
                
            dtype: data type (numpy-dtype); if array is in read-only
              mode, this and `frame_shape` are expected to be `None`,
              and the actual data type and frame shape of the target
              object will be ignored.
        
            frame_shape: shape of the data point; note that
              the actual shape of the resulting data set will contain
              one extra dimension upfront (i.e. the frame index).

            num_frames: Initial number of frames (i.e. the number of entries
             in the first dimension). `If this is 0, a dataset with 0 frames
             initially will be created, requiring `.push_frame()` to add data.
        
            *attrs: additional dataset attributes to set

        Return the dataset object.
        '''
        raise RuntimeError('not implemented')


    def _resize_dataset(self, dset, new_shape):
        '''
        Helper for `._ensure_frame()`.

        Essentially, this is the implementation-dependent part. Most backends
        will simply offer a `.resize()` call (this is true for HDF5 and Zarr),
        which is also the default implementation here.

        But others may require a more elaborate procedure (e.g. saving data,
        resizing, restoring).
        '''
        dset.resize(new_shape)


    def _reset_dataset(self, dset):
        '''
        Helper for `._ensure_frame()`.

        Erases all data points from a dataset.
        '''
        new_shape = [int(i) for i in (0, *(dset.shape[1:]))]
        self._resize_dataset(dset, new_shape)


    def _ensure_frame(self, dset, index):
        '''
        Ensures that `dset` has at least `index` points in its first dimension.

        Args:
            dset: Dataset object, dependent on the backend

            index: An integer counting the objects (starting with 1?)

        '''
        idiff = (index-dset.shape[0]) + 1

        if idiff == 1:
            shape_diff = 1
            
        elif idiff > 1:
            if '+' in self._frame_mode:
                shape_diff = idiff
            else:
                raise RefuseSparse(f'msg="Sparse writing requires top-up flag" '
                                   f'index={index} frames={dset.shape[0]} ')
            
        elif idiff <= 0:
            if 'a' in self._frame_mode: # append mode
                raise RefuseOverwrite(f'msg="Overwrite not permitted" index={index} '
                                      f'frames={dset.shape[0]} dataset="{dset.name}" '
                                      f'file="{self._recurse_obj_fname(dset)}"')

            elif 'o' in self._frame_mode or 'O' in self._frame_mode: # overwrite mode
                shape_diff = 0

            elif 'x' in self._frame_mode: # reset mode
                logger.info(f'msg="Resetting dataset" index={index} '
                            f'frames={dset.shape[0]} dataset="{dset.name}"')
                self._reset_dataset(dset)

                shape_diff = index+1

            else:
                raise RuntimeError(f'BUG: No shape diff for mode: {self._frame_mode}')

        if shape_diff > 0:
            new_shape = [int(i) for i in (dset.shape[0]+shape_diff, *(dset.shape[1:]))]
            logger.debug(f'msg="Resizing dataset" '
                         f'size={dset.shape[0]} '
                         f'index={index} diff={shape_diff} idiff={idiff} '
                         f'shape={new_shape}')
            self._resize_dataset(dset, new_shape)


    async def push_frame(self, index, **data):
        '''
        Received for every data point in a scan.

        Args:
            index: Integer to determine the position of the current
              data point in the overall (larger) dataset. This should
              always point to the next available index in the first
              dimention, i.e. if the existing dataset has dimension
              (37, N, M), with the current data point to become
              the 38th, then `index` needs to be 37 (counting starts
              at 0).

              Behavior if the index doesn't match the shape is defined
              by the class-global parameter `mode`.

              If this is `None`, the next free slot (i.e. current_size+1)
              is used, essentially making the operation always succeed
              as if "append" mode were active.

              If this is a negative number (e.g. -1) and the dataset
              doesn't exist, it is created according to shape
              specification of the input data, but with length 0
              (i.e. no points are copied).
              
            **data: dictionary from data name ("tag") to data arrays
        '''

        for key,container in data.items():
            dset_path = self._stor_dspath[key]
            grp, dset_only = self._split_dset_path(dset_path)
            self.ensure_group(grp)
            dset = self.ensure_dataset(dset_path,
                                       container.dtype,
                                       container.shape)

            if index is None:
                index = container.shape[0]

            if index >= 0:
                self._ensure_frame(dset, index)
                item = np_array(container)
                dset[index] = item


    def __getitem__(self, tag):
        '''
        Returns a handle to the object internally known as `tag`, according
        to the path spec specified at `__init__` time. This is generally
        supposed to be a dataset.
        '''
        dspath = self._stor_dspath[tag]
        obj = self.outer_object(tag)

        try:
            return obj[dspath]
        except KeyError:
            if not self.readonly:
                return self.ensure_dataset(dspath)
            raise



from camagick.executor import PipeExecutor
from camagick.storage.memory import ArrayAccess as MemoryArrayAccess

import tempfile, os, random, h5py, pytest, pprint, importlib
import numpy as np

from camagick.storage.hdf5 import ArrayAccess as H5ArrayAccess

for mod in ('astor',):
    name_here = mod[0].upper()+mod[1:]
    base = importlib.import_module(f'camagick.source.{mod}')
    globals()[name_here] = getattr(base, 'Processor')

@pytest.fixture()
def rndata():
    return np.random.rand(30, 36, 72)

@pytest.fixture()
def h5tmp(rndata):
    tmp = tempfile.NamedTemporaryFile(delete=False)
    tmp.close()
    with h5py.File(tmp.name, 'a') as h5:
        grp = h5.create_group('/group')
        grp.attrs['gmeta'] = 'group'
        data = grp.create_dataset(data=rndata, name='data')
        data.attrs['dmeta'] = 'dataset'
    yield tmp.name
    print('HDF5 temp file:', tmp.name)
    #os.unlink(tmp.name)


async def test_astor(h5tmp, rndata):

    ast = Astor(
        h5tmp,
        kind='hdf5',
        full='/group/data',
        part='/group/data[1:10:2,10:30]',
        dmeta='/group/data.attrs[dmeta]',
        gmeta='/group.attrs[gmeta]',
        dshape='/group/data.shape',
        dlist='/group.keys',
        root='.keys'
    )

    data = await ast()
    
    assert (data['full'] == rndata).all()
    assert (data['part'] == rndata[1:10:2,10:30]).all()
    assert data['dshape'] == rndata.shape
    assert data['dmeta'] == 'dataset'
    assert data['gmeta'] == 'group'
    assert data['dlist'][0] == 'data'

    #assert 'group' in h5


async def test_array_store():
    s1 = MemoryArrayAccess("testarray",
                           mode='a+',
                           d1="group/data1",
                           d2="group/sub/data2")

    async with s1:
        for i in range(10):
            await s1.push_frame(i,
                                d1=np.array([1*i, 2*i, 3*i]),
                                d2=np.array(3.14*i))

    pprint.pprint(s1._repo._storage)
    



from camagick.storage.base import ArrayAccess as ArrayAccessBase
from camagick.storage.base import RefuseIndex, RefuseOverwrite, RefuseSparse

import numpy as np
import h5py, logging, os, fsspec

logger = logging.getLogger(__name__)    

from contextlib import contextmanager 

class ArrayAccess(ArrayAccessBase):
    '''
    Implementation of a HDF5 file array access.
    '''

    #def __init__(self, outer=None, mode=None, **tags):
    #    '''
    #    Initializes an HDF5 array storage access wrapper.
    #
    #    Args:
    #
    #        outer: The outer part of the storage (i.e. path of the HDF file)
    #
    #        mode: Mode, as specified by the `ArrayAccess` base API. Note that
    #          this is _not_ the `h5py.File` mode!
    #
    #        tags: key=value access list, where "key" is the local
    #          handle, and "value" is the corresponding inner path.
    #          The value specified here is in addition to the "sdir"
    #          part (right?)
    #    '''
    #
    #    # HDF5 needs to have the root node named explicitly if
    #    # we want to access attributes like .keys or .attrs
    #    super().__init__(outer, mode, **tags)


    def _recurse_obj_fname(self, x):
        if hasattr(x, "filename"):
            return x.filename
        elif hasattr(x, "file") and x.file not in (x, None):
            return self._recurse_obj_fname(x.file)
        elif hasattr(x, "parent") and x.parent not in (x, None):
            return self._recurse_obj_fname(x.parent)
        return ''


    def __enter__(self):
        import contextlib
        self._exit_stack = contextlib.ExitStack()
        
        if self.readonly:
            self._file_obj = self._exit_stack.enter_context(
                fsspec.open(self.outer_path, mode='rb'))
            self._outer_object = self._exit_stack.enter_context(
                h5py.File(self._file_obj, 'r'))
            
        else:
            self._file_obj = self._exit_stack.enter_context(
                fsspec.open(self.outer_path, mode='rb+'))
            parent_dir = self._file_obj.fs._parent(self._file_obj.path)
            if not self._file_obj.fs.exists(parent_dir):
                self._file_obj.fs.mkdir(parent_dir)
            self._outer_object = self._exit_stack.enter_context(
                h5py.File(self._file_obj, 'a'))
            
        return self


    def outer_object(self, tag):
        # Ignore the tag -- this backend currently only supports single backend
        return self._outer_object
    

    def __exit__(self, *args):
        self._exit_stack.close()


    def ensure_group(self, name, attrs=None):
        # Makes sure `nodepath` exists in the current h5 file and is a data group
        nodepath = name
        subpath = ""
        subnode = self.outer_object(None)
        for current in nodepath.split('/'):

            if len(current) == 0:
                continue # nothing to do
            try:
                subnode = subnode[current]
                
            except KeyError:
                logger.info(f'msg="Creating node" file="{self._recurse_obj_fname(subnode)}" '
                            f'parent="{subnode.name}" '
                            f'name="{current}"')
                subnode = subnode.create_group(current)

            if attrs is not None:
                subnode.attrs.update(attrs)

        return subnode


    def _dtype_spec_to_dtype(self, ds):
        if isinstance(ds, str):
            try:            
                return {
                    'number': float,
                    'integer': int,
                    'array': float,
                }[ds]
            except KeyError:
                errmsg = f'tag="{key}" msg="Don\'t know how to save" dtype="{dt_spec}"'
                logger.error(errmsg)
                raise RuntimeError(errmsg)

        else:
            return ds


    def ensure_dataset(self, path, dtype=None, frame_shape=None, num_frames=0,
                       attrs=None):
        '''
        Received with the first data point, sets names, types etc.

        Args:
            name: name or path of the dataset        
            dtype: data type (numpy-dtype)
            point_shape: shape of the data point; note that
              the actual shape of the resulting data set will contain
              more dimension
        '''

        base = self.outer_object(None)

        if self.readonly:
            return base[path] if len(path) > 0 else base
        
        dshape = (0, *frame_shape) if frame_shape is not None else (0,)
        
        try:
            dset = base[path]

            if (dset.dtype != dtype):
                if  'O' not in self._frame_mode:
                    raise RuntimeError(f'msg="Dataset exists with different dtype"'
                                       f'have="{dset.dtype}" want="{dtype}"')
                else:
                    del base[path]
                    raise KeyError()

            if dset.shape[1:] != dshape[1:]:
                if  'O' not in self._frame_mode:                
                    raise RuntimeError(f'msg="Dataset exists with different shape"'
                                       f'have="{dset.shape}" want="{dshape}"')
                else:
                    del base[path]
                    raise KeyError()

        except KeyError:
            logger.info(f'msg="Creating dataset" file="{base.file.filename}" name="{path}" '
                        f'shape="{dshape}" dtype={dtype}')
            dset = base.create_dataset(path, dshape, maxshape=(None, *dshape[1:]),
                                       dtype=dtype, compression='lzf')

        if attrs is not None:
            dset.attrs.update(attrs)

        return dset



#!/usr/bin/python3

import logging, time, math, random, asyncio

from camagick.processor import ProcessorBase, SourceBase
from camagick.probe import opcast
from asteval import Interpreter

logger = logging.getLogger(__name__)

class Processor(SourceBase):
    '''
    Produces data from scratch by evaluating mathematical expressions.

    The eval expressions are parsed by the `asteval` module , providing
    a useful subset of the Python language focused on mathematical
    expresssions, mostly in the form of `numpy` functions and basic
    arithmetic operations (linspace, sin, ones, random, ...).

    This is a more powerful and easy-to-use data source than the
    "scratch" module, and actually in almost every way superior (except
    for the fact that `asteval` doesn't have a random number generator).
    But with great power comes great responsibility, as the underlying
    `asteval` module might allow for expressions that don't strictly
    result in array data -- which would break your CAspy flow.
    '''

    def __init__(self, **tags):
        '''
        Args:
            **tags: map of data name -> eval expression to generate
              data from.
        '''

        self._tags = tags
        self._inter = Interpreter()
        super().__init__()


    async def __call__(self, data=None, context=None):
        d = {}

        self._inter.symtable.update(context)

        for tag,code in self._tags.items():
            ev = self._inter(code)
            if ev is None:
                raise RuntimeError(f'msg="Asteval error" detail="{self._inter.error}" '
                                   f'tag="{tag}"')
                
            d[tag] = ev

        return d



from camagick.reader import Receiver

from caproto.sync import client as ca_client
from caproto.asyncio.client import Context
from caproto import CaprotoTimeoutError
import numpy as np
import xarray as xr

import asyncio

import logging
logger = logging.getLogger(__name__)


def extract_data(response, pvName=None, others=None, auto_xarray=False):
    '''
    Extracts "useful" data out of a CAproto response telegram.
    '''
    pass


class EpicsReceiver(Receiver):
    ## Base class for all EPICS-based receivers -- essentially
    ## just manages the CAproto client context and data extraction
    async def init(self):
        await super().init()
        self.ctx = self.ctx = Context()
        self.data_pvs = await self.ctx.get_pvs(*[v for v in self._tokens])
        

    async def shutdown(self):
        await self.ctx.disconnect()


    def unpack(self, token, dataobj, context):
        
        auto_xarray=False
        others=context
        pvName=token
        response=dataobj

        if response is None:
            return None

        if others is None:
            others = {}

        # Channel types can be: CHAR, DOUBLE, FLOAT, STRING, ENUM, LONG, INT.
        # The intention is to get an automatic useful native Python data type,
        # scalar or array. This means different things for different data
        # types.
        # In addition, we implement some heuristics to decorate waveforms
        # (== arrays) if our obscure array markers are present (shape, dimensions,
        # axis scaling -- to be documented ;-) )

        if response.data_type in (ca_client.ChannelType.STRING,):
            return response.data[0].decode('utf-8')

        elif response.data_type in (ca_client.ChannelType.DOUBLE,
                                    ca_client.ChannelType.FLOAT,
                                    ca_client.ChannelType.LONG,
                                    ca_client.ChannelType.INT,
                                    ca_client.ChannelType.ENUM):

            if len(response.data) == 1:
                return response.data[0]

            if not pvName or not pvName.endswith("_SIGNAL"):
                return response.data

            # If we have an array and it ends on _SIGNAL, we also try to
            # load _OFFSET and _DELTA for intrinsic scaling information
            o_name = pvName.replace("_SIGNAL", "_OFFSET")
            d_name = pvName.replace("_SIGNAL", "_DELTA")

            if o_name in others:
                offs = extract_data(others.get(o_name, 0))
            else:
                offs = 0

            if d_name in others:
                dlta = extract_data(others.get(d_name, 1))
            else:
                dlta = 1

            try:
                axis = offs+np.array(range(len(response.data)))*dlta
            except TypeError as e:
                # This happens when not all the data (e.g. `dlta` or `offs`
                # has arrived yet.
                axis = np.array([np.nan] * len(response.data))

            if auto_xarray:
                return DataArray(data=response.data, dims=["x"], coords=[axis])
            else:
                return response.data

        elif response.data_type in (ca_client.ChannelType.CHAR,):
            # This is a string -- return as such (ACHTUNG, this will break
            # "everything is a numpy array" data philosophy -- but we _want_
            # this. There's no other way to discern a uint8 from a string
            # later on.
            p = response.data
            s = bytes(p).decode('utf-8')
            return s

        elif response.data_type in (ca_client.ChannelType.STRING,):
            # This is a string -- return as such (ACHTUNG, this will break
            # "everything is a numpy array" data philosophy -- but we _want_
            # this. There's no other way to discern a uint8 from a string
            # later on.
            #print('have', repsonse.data)
            return response.data


        # else: how to handle ENUM / CHAR?

        else:
            logger.warning ("Unhandled data type: %r" % (response.data_type,))
            return response.data[0]

    
class SubscriptionReceiver(EpicsReceiver):
    ## Async subscription-based implementation of "Receiver"
    async def init(self):
        await super().init()        
        self._ca_subscriptions = {}
        for d in self.data_pvs:
            logger.info(f'msg="Subscribing" pv="{d.name}"')
            self._ca_subscriptions[d.name] = d.subscribe()
            self._ca_subscriptions[d.name].add_callback(self._process_subscription)
        logger.debug(f'msg="CA client contex" ctx={self.ctx}')

    def _process_subscription(self, sub, data):
        self._process(sub.pv.name, data)


class PollingReceiver(EpicsReceiver):
    ## Polling-based receiver; essentially just installs a polling task
    ## with a (fairly slow) period and repeatedly tries to capture
    ## the PV values by calling .read().
    
    async def init(self):
        await super().init()
        poll_period = 0.2 # seconds
        self._poll_task = asyncio.create_task(self._poll(poll_period))

    async def _poll(self, poll_period):
        async def _read_and_process(pvobj, oldval):
            self._process(pvobj.name, (await pvobj.read()))
        self._old_data = None

        try:
            while True:
                data = await asyncio.gather(
                    asyncio.sleep(poll_period),
                    *[ _read_and_process(pv, None) for pv in self.data_pvs ],
                    return_exceptions = False
                )
                self._old_data = data
        except CaprotoTimeoutError as e:
            logger.warning(f'msg="CAproto timeout" detail="{e}"')
            pass
        except Exception as e:
            logger.error(f'msg="CA polling failed, detail follows" error={type(e)}')
            logger.error(e)
            logger.error(traceback.format_exc())
            raise
        finally:
            pass


    async def shutdown(self):
        try:
            self._poll_task.cancel()
            await self._poll_task
        except asyncio.CancelledError: pass



from camagick.processor import ProcessorBase
import numpy as np
import logging
from asteval import Interpreter

logger = logging.getLogger(__name__)

def _parse_slice(s, inter=None):

    if inter is None:
        inter = Interpreter()
    
    parts = s.split(':')
    
    if len(parts)==1:
        return None if len(parts[0])==0 \
            else inter(parts[0], raise_errors=True)

    if len(parts) > 0:
        sparam = [int(inter(p, raise_errors=True) if p not in ("None", None) else None) \
                  if len(p)>0 else None for p in parts]
        return slice(*sparam)


class Processor(ProcessorBase):    
    '''
    Slices, a.k.a. selects, a subset of each dataset.
    '''

    def __init__(self, *rois):
        '''
        Iteratively selects regions of interests from data.

        Args:
            *rois: each unnamed parameter is either a `slice()` object (if used in API
              mode), a string of the format "start:stop:step" or "spec", or `None`.
              If it's a string, a slice object is constructed from the parsed string
              values using `asteval.Interpreter()` from the `lmfit` people -- meaning
              that you actually have access to a number of arithmetical operations
              and the context data as sy,bols. If it's `None`, it will be treated
              like `numpy`'s `None` or `newaxis`: a new axis is produced.
        
              As a convenience, if there's only one dimensional specification (i.e.
              one parameter), which is a string and which contains "," it treated
              as a complete specification of the form "dim,dim,dim".
        '''

        self._roi_spec = rois


    def _parse_roi(self, rois, ctx):

        # Splitting of single-parameter specification
        if len(rois)==1 and isinstance(rois[0], str):
            r = rois[0].split(',')
            rois = r

        # ROR params, one for each dimension.
        roi_params = []

        inter = Interpreter()
        inter.symtable.update(ctx)
        
        for r in rois:
            if isinstance(r, str):
                try:
                    o = inter(r, show_errors=False, raise_errors=True)
                    obj = int(o) if o not in (None,) else o
                except SyntaxError:
                    try:
                        obj = _parse_slice(r, inter)
                    except SyntaxError as e:
                        logger.error(f'msg="Error parsing ROI dimension" spec="{r}" '
                                     f'detail="{e}"')
                        raise
                roi_params.append(obj)
            elif hasattr(r, "__len__"):
                raise RuntimeError(f"Don't know how to deal with {r}")
            else:
                roi_params.append(r)

        return roi_params        
    
    async def __call__(self, data=None, context=None):
        ''' Executes the ROI selection. Context is ignored. '''
        try:
            roi_params = self._parse_roi(self._roi_spec, context or {})
            return {
                key:val[*roi_params] for key,val in data.items()
            }
        except Exception as e:
            logger.error(f"roi={self._roi_spec}")
            raise



from camagick.processor import ProcessorBase, SinkBase

import multiprocessing as mp
import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib.widgets import Cursor, RectangleSelector
from matplotlib.patches import Rectangle, Circle
from matplotlib.colors import LogNorm
import numpy as np

import traceback, os

from camagick.executor import QuitApplication

import time, math, logging, concurrent, asyncio, pickle, asteval

logger = logging.getLogger(__name__)


def resample_array(name, data, num_bins):
    if len(data) > num_bins:
        slices = np.linspace(0, data.shape[0], num_bins+1, True).astype(int)
        counts = np.diff(slices)
        mean = np.add.reduceat(data, slices[:-1]) / counts
        return mean
    else:
        return data

class MatplotlibDisplay:
    '''
    Stupidly simple class that uses Matplotlib to visualize numpy arrays.
    '''
    def __init__(self, rows=None, cols=None, flip_lr=True, transpose=False, sample_max=2056):
        self.flip_lr = flip_lr
        self.transpose = transpose
        self._rows = rows
        self._cols = cols
        self._sample_max = sample_max
        self.figure = plt.figure()
        self.bgcolor = (0.3, 0.3, 0.3)
        self.figure.patch.set_facecolor(self.bgcolor)

        self.panels = {}  # plots (lines/image) objects, per-tag
        self.axesobj = {} # actual axes objects _per_ _layout_
        self.axes = {}    # axes (subplot) references, organized per-tag
        self.markers = {} # marker objects (rect etc) to display, per-tag

        self.set_expected(1)
        

    def _default_rows_cols(self, num_panels, rows, cols):
        if rows is not None:
            self._rows = rows

        if cols is not None:
            self._cols = cols

        rows = self._rows if self._rows is not None else \
            int(math.ceil(math.sqrt(num_panels)))

        if rows == 0:
            logger.error(f'msg="ACHTUNG, no data"')
            rows += 1
        
        cols = self._cols if cols is not None else \
            int(math.ceil(num_panels/rows))

        return rows, cols


    def init_display(self, panel_name, pspec=None):
        '''
        Initializes display for a specific panel.
        Uses panel specification as in `pspec`, if not `None`.
        `num_total` is only used to determine a default subplot layout/configuration
        in the absence of a `pspec.sub` specification.
        '''

        num_total = self._expected_num_panels
        
        # default layout, if pspec doesn't exist
        rows, cols = self._default_rows_cols(num_total, None, None)

        # how many panels did we already place
        have = len(self.axes)

        #logger.info(f'geometry={rows}x{cols} displays={self.panelNames}')
        #self.axes = { k: self.figure.add_subplot(rows, cols, i+1) \
        #              for i,k in enumerate(self.panelNames) }

        if pspec is not None and pspec.subplot is not None:
            # Need to transform the subspec to the 3-digit-integer format,
            # because we want to use this as a key in .axesobj.
            if hasattr(pspec.subplot.sub, '__len__'):
                r, c, h = pspec.subplot.sub
                subspec = r*100+c*10+h
            else:
                subspec = pspec.subplot.sub
            subargs = pspec.subplot.kwargs
        else:
            subspec = rows*100 + cols*10 + have+1
            subargs = {}

        if subspec not in self.axesobj:
            ax = self.axesobj[subspec] = self.figure.add_subplot(subspec, **subargs)
            ax.set_facecolor(self.bgcolor)            
            logger.error(f'msg="Adding subplot" panel={panel_name} spec={subspec} sub={ax}')
        else:
            ax = self.axesobj[subspec]
            #ax.legend()

        if panel_name not in self.axes:
            logger.error(f'msg="Attibuting panel to subplot" panel={panel_name} sub={ax}')
            self.axes[panel_name] = ax

        self.markers[panel_name] = {}
        
            #RectangleSelector(
            #    ax, useblit=True,
            #    interactive=True,
            #    props={'fill': False},
            #    drag_from_anywhere=True,
            #    #ignore_event_outside=True
            #)
        #self.last_update = {k:time.time() for k in self.panelNames }


    def _plot_norm_2d(self, img):
        mi = img.min()
        ma = img.max()
        return mi, ma, LogNorm(vmin=mi if mi > 0 else 1e-5,
                               vmax=ma if ma > 0 and ma > mi else 1)
    
        
        
    def _plot_2d(self, name, data, *kw):
        ax = self.axes[name]
        ax.set_title(name)

        if self.flip_lr:
            img = data[:,::-1]
        else:
            img = data

        if self.transpose:
            img = np.transpose(img)
        
        self._2d_norm = self._plot_norm_2d(img)

        return ax.imshow(img, norm=self._2d_norm[2], *kw)


    def _plot_1d(self, name, data, *kw):        
        d = resample_array(name, data, self._sample_max)

        if len(data) != len(d):
            logger.info(f'tag={name} msg="Data was rebinned" '
                        f'full_shape={data.shape} '
                        f'bin_shape={d.shape}')        
        
        ax = self.axes[name]
        x = ax.plot(d,*kw)

        # Update legend to include all plots for this panel
        labels = [k for k,v in filter(lambda x: x[1] == ax, self.axes.items()) ]
        ax.legend(labels)

        return x


    def _update_1d(self, name, data, pspec=None, **kw):

        if (pspec is not None):
            axis = pspec.axes[0]
        else:
            axis = np.array(range(len(data)))
        
        if len(data) > self._sample_max:
            d = resample_array(name, data, self._sample_max)
        else:
            d = data

        x2 = resample_array(name, axis, self._sample_max)

        line = self.panels[name][0]
        line.set_xdata(x2)
        line.set_ydata(d)
        line.axes.relim()
        line.axes.autoscale_view()


    def _update_2d(self, name, data, pspec=None, **kw):
        if self.flip_lr:
            img = data[:,::-1]
        else:
            img = data

        if self.transpose:
            img = np.transpose(img)

        self.panels[name].set_data(img)

        if (pspec is not None) and \
           (hasattr(pspec, "cursor")) and \
           (pspec.cursor is not None):
            for ci,cursor in enumerate(pspec.cursor):
                try:
                    patch = self.markers[name].get(ci)
                    if len(cursor) > 2:
                        if patch is None:
                            patch = self.markers[name][ci] = \
                                Rectangle((cursor[0], cursor[1]), cursor[2], cursor[3],
                                          edgecolor='red', facecolor=(0.0, 0.0, 0.0, 0.0))
                            self.axes[name].add_patch(patch)
                        else:
                            patch.set(x=cursor[0], y=cursor[1], width=cursor[2], height=cursor[3])
                    else:
                        if patch is None:
                            patch = self.markers[name][ci] = Circle((cursor[0], cursor[1]))
                        else:
                            patch.set(center=(cursor[0], cursor[1]))
                except KeyError as e:
                    logger.error(f'msg="Cannot set marker" name="{name}" reason="{e}"')
                    print(self.markers.keys())


        mi, ma, norm = self._plot_norm_2d(img)
        if mi < self._2d_norm[0] or ma > self._2d_norm[1]:
            logger.info(f'msg="Updating norm" tag="{name}" vmin="{mi}", vmax="{ma}"')
            self.panels[name].set_norm(norm)
            self._2d_norm = (mi, ma, norm)


    def _panel_for_data(self, name, data, pspec):
        '''
        Returns the panel for data set `name`
        '''
        dims = len(data.shape)
        plot_func = getattr(self, f"_plot_{dims}d")
        plot_kwargs = {
            1: {},
            2: {}
        }[dims]

        if data is None:
            data = np.ndarray([2]*dims)

        pan = self.panels.get(name)
        if pan is None:
            try:
                self.init_display(name, pspec)
                self.panels[name] = pan = plot_func(name, data, **plot_kwargs)
            except Exception as e:
                traceback.print_exc(e)
                raise

        return plt


    def set_expected(self, num_panels):
        self._expected_num_panels = num_panels
        

    def update(self, panel, data=None, pspec=None):
        
        if data is None:
            return
        
        try:
            dims = len(data.shape)
            ax = self._panel_for_data(panel, data, pspec)
        except KeyError:
            logging.error("%s: no such display panel" % panel)
            return

        getattr(self, f"_update_{dims}d")(panel, data, pspec)
        self.figure.canvas.draw_idle()


    def run(self):
        plt.show()


def _send(pipe, msg):
    try:
        pipe.send(msg)
    except Exception as e:
        logger.error(f'msg="Failed to send data to plotter" detail="{e}"')
        raise
        
        
class ProcessPlotter:

    def __init__(self, panels=None):
        self._panspec = panels if panels is not None else {}
        self.send_pipe, self.recv_pipe = mp.Pipe()
        self.plotter_process = mp.Process(
            target=self, args=(self.recv_pipe,), daemon=True
        )
        self.plotter_process.start()


    async def plot(self, data, context):
        '''
        Asynchronously sends data to the plotting (matplotlib) process.
        '''
        with concurrent.futures.ThreadPoolExecutor() as pool:
            try:
                loop = asyncio.get_running_loop()
                task = loop.run_in_executor(pool, _send, self.send_pipe,
                                        {'data': data, 'ctx': context})
                await task
            except Exception as e:
                logger.error(f'msg="Plotter data feed" error="{e}"')
                raise

    
    def terminate(self):
        self.send_pipe.send(None)
        plt.close('all')

        
    def _eval_plotspec(self, ps, data, ctx):
        # Plotspec string: [pos, {kwargs}], xaxis, yaxis, cursorAt, cursorSize]
        from asteval import Interpreter
        inter = Interpreter()
        inter.symtable.update(ctx)
        from collections import namedtuple
        param_names = ['subplot', 'axes', 'cursor']
        PlotSpec = namedtuple('plotspec', param_names)
        have_args = inter(ps, raise_errors=True)
        if have_args is None:
            return None
        have = list(have_args)
        need =  [None]*(len(param_names)-len(have_args))

        interm = list(have_args) + need
        pspec = PlotSpec(*interm)

        # Complete missing parts.

        # Subplot must either be an integer 'xyz', a tuple (x,y,z),
        # or a tuple and a dictionary (subplot, {kwargs})
        if type(pspec.subplot) == int \
           or ((hasattr(pspec.subplot, "__len__") and len(pspec.subplot)) == 3) \
           or (pspec.subplot is None):
            pspec_subplot = namedtuple('PlotSpecSubplot', ['sub', 'kwargs'])(pspec.subplot, {})
        elif len(pspec.subplot) != 2:
            raise RuntimeError(f'Bad subplot spec \'{pspec.subplot}\'')
        else:
            if type(pspec.subplot[0]) == int:
                pspec_subplot = namedtuple('PlotSpecSubplot', ['sub', 'kwargs'])(pspec.subplot[0], pspec.subplot[1])
            else:
                pspec_subplot = namedtuple('PlotSpecSubplot', ['sub', 'kwargs'])(*pspec.subplot)


        is_standalone = lambda x: len(data.shape)==1 \
            and (x is not None) and \
            (hasattr(x, "__len__") and len(x)) != 1

        # Axes spec is one item per axis -- except if the dimension
        if is_standalone(pspec.axes):
            pspec_axes = (pspec.axes,)
        else:
            pspec_axes = pspec.axes
        # As a convenience, if the axis data is a (a,b) tuple, we
        # interpret it as a (delta, offset) specification and expand
        # to a linspace
        if pspec_axes is not None:
            pspec_axes = tuple([
                (np.linspace(start=p[0], stop=p[0]+p[1]*ds, num=ds) \
                 if p is not None and hasattr(p, "__len__") and len(p) == 2 \
                 else p) \
                for p,ds in zip(pspec_axes,data.shape)
            ])

        # Cursor is a tuple/list with item per dimension (or a single item
        # if the dimenson is 1).
        # Each item is, in turn, either a single- or a double-number.
        # Single-number is the position of the cursor (spanning from left
        # to right), double-number is position and size.
        if is_standalone(pspec.cursor):
            pspec_cursor = (pspec.cursor,)
        else:
            pspec_cursor = pspec.cursor
        if pspec_cursor is not None:
            pspec_cursor = tuple([
                    (p,0) if len(p)==0 else tuple([x for x in p]) \
                    for p in pspec_cursor
            ])

        return PlotSpec(pspec_subplot, pspec_axes, pspec_cursor)

    
    def callback(self):
        data = {}
        m0 = time.time()
        while self.pipe.poll():
            incoming = self.pipe.recv()
            if incoming is None or incoming['data'] is None:
                self.terminate()
                return False
            
            data = incoming['data']
            context = incoming['ctx'] or {}

        if len(data) > 0:
            self.display.set_expected(len(data))

        for dname,dvalues in data.items():
            try:
                pspec = self._panspec.get(dname, '')
                ps = self._eval_plotspec(pspec, dvalues, context)
                #logger.debug(f'tag={dname} subplot={ps.subplot} axes={ps.axes} cursor={ps.cursor}"')
            except Exception as e:
                logger.error(f'msg="Bad plot spec" tag=dname from=\'pspec\' reason="{e}"')
    
            
        #if (not self.display.is_initialized):            
        #    if len(data)>0:
        #        self.display.init_display(*[k for k in data.keys()])

        for name,val in data.items():
            self.display.update(name, val, ps)
            
        return True


    def __call__(self, pipe=None, period=0.03):
        self.pipe = pipe or self.recv_pipe
        self.display = MatplotlibDisplay(transpose=True, flip_lr=False)
        intr = int(period*1000)
        logger.info(f'msg="Matplotlib subprocess" refresh={intr}ms pid={os.getpid()}')
        timer = self.display.figure.canvas.new_timer(interval=intr)
        timer.add_callback(self.callback)
        timer.start()
        
        self.display.run()


class Processor(SinkBase):
    '''
    Plots data as lines (1D arrays) or images (2D arrays).
    '''
    
    def __init__(self, **spec):
        '''
        Args:
            **spec: Dictionary with data tags as keys, and plot spec
              strings as values. The plot spec strings are specified
              as a string `pos,spec,xaxis,yaxis`
              the following meaning:
        
              - pos: a subplot position spec compatible with the arguments
                passed to a call to `matplotlib.pyplot.subplot()`, i.e.
                "221" or "[2,2,1]".

              - spec: keyword arguments for subplot spec, to be passed
                to `matplotlib.pyplot.subplot()`, enclosed in `dict()`
                brackets `{}`, e.g.:  `{"polar":True}` or `{"sharex":False}`

              - xaxis: exxpression to calculate the X axis
                
        '''

        # Another tentative format: `"[sub[:x[:y]]][#fx[:fy][,tx[:ty]]"`
        self._panspec = spec


    async def startup(self):
        self.plotter = ProcessPlotter(self._panspec)


    async def shutdown(self):
        self.plotter.terminate()

    
    async def __call__(self, data=None, context=None):
        try:
            await self.plotter.plot(data, context)

            #return { k:data[k] for k in \
            #         filter(lambda x: x not in data, data.keys()) }
            
        except BrokenPipeError:
            logger.info(f'msg="Display closed by user"')
            raise QuitApplication(f'msg="Display closed by user"')
